function [x, j, res_norm, res_hist] = cgnr(A, x, r, tol, maxiter)
%CGNR Conjugate Gradient method on the Normal equations
%
%  This function solves the matrix system A*x = b, by applying the
%  Conjugate Gradient (CG) method on the normal equations A'*A*x = A'*b.
%  Author      :  Reijer Idema (http://www.reijeridema.nl)
%  Last edited :  15 September 2008
%
%  Parameters
%     A        :  coefficient matrix
%     x        :  initial solution x_0
%     r        :  initial residual r_0 = b - A*x_0
%     tol      :  stop criterion on the relative residual norm
%     maxiter  :  stop criterion on the number of iterations
%
%  Output
%     x        :  approximate solution to the system A*x = b
%     j        :  iteration count
%     res_norm :  residual norm
%     res_hist :  residual norm history through all iterations
%
%  Complexity
%     matrix-vector     :  2
%     innter products   :  3
%     vector updates    :  3
%
%  Memory requirement
%     matrix memory     :  1
%     vector memory     :  5 (not counting res_hist)

%  Note that CGNR requires that matrix vector products with the
%  transpose of A can be computed efficiently.

%% Initialize variables
j = 0;

z = A'*r;                                       % z_0 is the residual of the normal equations
s = z;                                          % s_0 = z_0
zip = z'*z;                                     % normal equations residual inner product

norm_r0 = norm(r,2);                            % norm_r0 = ||r_0||
res_norm = norm_r0;                             % res_norm = ||r_0||
res_hist = zeros(maxiter+1, 1);                 % residual history preallocation
res_hist(1) = res_norm;

%% MRS3 iteration loop
while (res_norm/norm_r0 > tol && j<maxiter)

   j = j+1;
   v = A*s;                                     % v_j = A*s_j

   alpha = zip/(v'*v);                          % alpha = (z_j'*z_j)/(v_j'*v_j)
   x = x + alpha*s;                             % update solution approximation
   r = r - alpha*v;                             % r is the residual of the original system
   z = A'*r;                                    % z is the residual of the normal equations

   zip_old = zip;
   zip = z'*z;
   beta = zip/zip_old;                          % beta = (z_{j+1}'*z_{j+1})/(z_j'*z_j)
   s = z + beta*s;                              % s_{j+1} = z_{j+1} + beta*s_j

   res_norm = norm(r,2);                        % residual norm of the original system
   res_hist(j+1) = res_norm;

end

%% End
